# Generated by Django 4.2.11 on 2024-05-27 11:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0026_create_unaccent_full_text_search_config"),
    ]

    operations = [
        migrations.AddField(
            model_name="worker",
            name="private_key",
            field=models.BinaryField(default=b"", max_length=40),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="worker",
            name="public_key",
            field=models.BinaryField(default=b"", max_length=40),
            preserve_default=False,
        ),
    ]
