# Generated by Django 4.2.16 on 2024-11-21 16:23

import django.db.models.manager
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0031_remove_annotationvalue_uniquetogether_entity_field_and_more"),
    ]

    operations = [
        migrations.AlterModelManagers(
            name="process",
            managers=[
                ("unversioned_objects", django.db.models.manager.Manager()),
            ],
        ),
    ]
